/*
 * Light-weight Reliable Multicast
 *
 * lrm.c - Main lrm activity handling.
 * Authors: Yoann Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "lrm_list.h"
#include "lrm.h"
#include "lrm_timers.h"
#include "lrm_api.h"
#include "lrm_messages.h"

int lrm_debug_ = 0;

#ifdef LRM_MEM_DEBUG
void * talloc(int size, char * file, int line, lrm_alloc_f alloc) {
  void * m;
  lrm_err("[lrm_mem %s:%d] malloc(%d)", file, line, size);
  fflush(stderr);
  m = alloc(size);
  lrm_err(" = %d\n", (int) m);
  return m;
}
#endif

void lrm_err(const char * format, ...) {
  va_list ap;

  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
}

char * lrm_host(unsigned long addr) {
  static char buf[255];
  struct hostent *host;
  struct in_addr inp;

  inp.s_addr = addr;
  host = gethostbyaddr((char *) &inp, sizeof(struct in_addr), AF_INET);
  if (host && host->h_name) {
    if (strlen(host->h_name)>230)
      host->h_name[230]='\0';
    sprintf(buf, "%s [%s]", host->h_name, inet_ntoa(inp));
  } else
    sprintf(buf, "%s", inet_ntoa(inp));
  
  return buf;
}

char * lrm_addr(unsigned long addr) {
  static char buf[255];
  struct in_addr inp;

  inp.s_addr = addr;
  sprintf(buf, "%s", inet_ntoa(inp));
  return buf;
}

void lrm_free_user(struct lrm_session *s, struct lrm_user * u) {
  lrm_debug (("removing uid %8x (%s) from user-list.\n", u->uid, lrm_host(u->addr)));
  if (u->last_ack_as_receiver != u->last_delivered)
    lrm_send_ack(s, u);
  lrm_free_recvd_messages(s, u);
  free(u);
}

struct lrm_user * lrm_new_user(struct lrm_session *s, lrm_uid uid, unsigned long addr) {
  struct lrm_user * us;

  us = (struct lrm_user *) malloc(sizeof(struct lrm_user));
  if (!us) return NULL;
  
  lrm_debug (("adding uid %8x (%s) to user-list.\n", uid, lrm_host(addr)));
  
  us->messages = lrm_list_new();
  if (!us->messages)
    lrm_memerr ("lrm_list_new");
  us->uid = uid;
  us->addr = addr;
  us->last_delivered = 0;
  us->last_ack_as_receiver = 0;
  us->last_ack_as_sender = s->seqno - 1;	    
  us->ping = LRM_RESET_PING;
  us->wait = s->wait_for_resent;
  
  if (lrm_list_insert_last(s->users, (void *) us) == LRM_FAILURE)
    lrm_memerr ("lrm_list_insert_last");
  return us;
}

struct lrm_user * lrm_get_user(struct lrm_session *s, lrm_uid uid) {
  lrm_iterator u;

  for (u=lrm_list_begin(s->users); u!=NULL; u=lrm_list_next(u))
    if (((struct lrm_user *) lrm_list_content(u))->uid == uid) break;

  if (u!=NULL)
    return (struct lrm_user *) lrm_list_content(u);
  else
    return NULL;
}

void  lrm_del_user(struct lrm_session *s, lrm_uid uid) {
  lrm_iterator u;
  struct lrm_user * us;
  
  for (u=lrm_list_begin(s->users); u!=NULL; u=lrm_list_next(u))
    if (((struct lrm_user *) lrm_list_content(u))->uid == uid) break;
  if (!u) return;
  
  us = (struct lrm_user *) lrm_list_content(u);
  lrm_free_user(s, us);
  lrm_list_delete(s->users, u);
}

void lrm_read_socket(struct lrm_session *s) {
  struct sockaddr_in remoteExtremity;
  char buffer [LRM_MAX_PACKET_SIZE];
  int len;
  int socklen;
  lrm_key lrmkey;
  lrm_type type;
  lrm_uid uid;
  lrm_seqno seqno;
  unsigned long addr;
  lrm_seqno ack;
  int deleted;
  struct lrm_user * user;
  lrm_iterator u, m;
#ifdef LRM_DEBUG
  char * address;
#endif
  void * message;
  int message_length;
  struct lrm_recvd_message * recvd_message;

  socklen = sizeof(struct sockaddr);
  len = recvfrom(s->sock, buffer, LRM_MAX_PACKET_SIZE, 0,
		 (struct sockaddr *) &remoteExtremity, &socklen);
  
  addr = remoteExtremity.sin_addr.s_addr;
#ifdef LRM_DEBUG
  address = inet_ntoa(remoteExtremity.sin_addr);
#endif
  
  if (len < LRM_MIN_PACKET_SIZE) {
    lrm_debug (("received too short message from %s.\n", address));
    return;
  }

  memcpy (&lrmkey, buffer, sizeof(lrm_key));
  lrmkey = ntohl (lrmkey);
  if (lrmkey != LRM_KEY) {
    lrm_debug (("received message with invalid key from %s.\n", address));
    return;
  }

#ifdef LRM_DEBUG
  if (s->drop_probability!=0 && s->drop_probability<=1)
    if ((double) rand () < ((double) RAND_MAX * s->drop_probability)) {
      lrm_debug(("dropping message of %d bytes from %s.\n", len, address));
      return;
    }
#endif
  
  memcpy(&uid, buffer + sizeof(lrm_key),
	 sizeof(lrm_uid));
  memcpy(&type, buffer + sizeof(lrm_key) + sizeof(lrm_uid),
	 sizeof(lrm_type));
  memcpy(&seqno, buffer + sizeof(lrm_key) + sizeof(lrm_uid) + sizeof(lrm_type),
	 sizeof(lrm_seqno));
  
  type = ntohs (type);
  uid = ntohl (uid);
  seqno = ntohl (seqno);

  user = lrm_get_user(s, uid);
  if (!user && (type != LRM_JOIN)) {
    if ((s->options & LRM_STRICT) && (type != LRM_PING) && (type != LRM_PONG)) {
      lrm_debug (("message %d from %8x (%s, %u) who is unknown.\n", type, uid, address, seqno)) ;
    } else {
      user = lrm_new_user(s, uid, addr);
      if (!user)
	lrm_memerr ("lrm_new_user");
      else
	user->last_delivered = user->last_ack_as_receiver = seqno;
    }
  }

  if (!user && (type != LRM_JOIN)) return;
  if (user) user->ping = LRM_RESET_PING;

  switch (type) {
    case LRM_DIFFUSION:
      if (len < LRM_MIN_PACKET_SIZE + sizeof(lrm_seqno)) {
	lrm_debug (("received diffusion message with wrong data from %8x (%s).\n", uid, address));
	return;
      }

      message_length = len - LRM_MIN_PACKET_SIZE - sizeof(lrm_seqno); 

      recvd_message = (struct lrm_recvd_message *) malloc(sizeof(struct lrm_recvd_message));
      if (!recvd_message) {
	lrm_memerr ("lrm_read_socket");
	return;
      }
      
      memcpy(&recvd_message->seqno, buffer + LRM_MIN_PACKET_SIZE, sizeof(lrm_seqno));
      recvd_message->seqno = ntohl (recvd_message->seqno);
      recvd_message->addr = addr;
      recvd_message->len = message_length;
      if (message_length) {
	recvd_message->data = malloc(message_length);
	if (!recvd_message->data) {
	  lrm_memerr ("lrm_read_socket");
	  return ;
	}
	memcpy(recvd_message->data, buffer + LRM_MIN_PACKET_SIZE + sizeof(lrm_seqno),
	       message_length);
      } else
	recvd_message->data = NULL;
      
      lrm_debug (("received message %d (%d bytes) from %8x (%s, %u), expecting %d.\n",
		  recvd_message->seqno, message_length, uid, address, seqno, user->last_delivered+1));
      
      /* 
	if (user->last_delivered == 0) 
	  user->last_delivered = (recvd_message->seqno) - 1;
      */
      
      if (recvd_message->seqno <= user->last_delivered) {
	lrm_debug (("received already processed message from %8x (%s, %u). ignored.\n", uid, address, seqno));
	lrm_free_recvd_message(recvd_message);
	return;
      }

      if (recvd_message->seqno == user->last_delivered+1) {
	lrm_give_message(s, user, recvd_message);
	lrm_give_user_queue(s, user);
      } else {
	lrm_iterator it = lrm_list_begin (user->messages);
	if (it != NULL) {
	  while ((((struct lrm_recvd_message *) lrm_list_content (it))->seqno < recvd_message->seqno) && (lrm_list_next (it) != NULL))
	    it = lrm_list_next(it);
	  if (((struct lrm_recvd_message *) lrm_list_content (it))->seqno == recvd_message->seqno) {
	    lrm_debug (("message already saved in list.\n", address));
	    lrm_free_recvd_message(recvd_message);
	  } else
	    if (((struct lrm_recvd_message *) lrm_list_content (it))->seqno > recvd_message->seqno) {
	      if (lrm_list_insert_before (user->messages, it, (void *) recvd_message) == LRM_FAILURE) {
		lrm_memerr ("lrm_list_insert_before");
		lrm_free_recvd_message(recvd_message);
	      }
	    } else
	      if (lrm_list_next (it) == NULL) {
		if (lrm_list_insert_last (user->messages, recvd_message) == LRM_FAILURE) {
		  lrm_memerr ("lrm_list_insert_last");
		  lrm_free_recvd_message(recvd_message);
		}
	      }
	} else {
	  if (lrm_list_insert_last (user->messages, recvd_message) == LRM_FAILURE) {
	      lrm_memerr ("lrm_list_insert_last");
	      lrm_free_recvd_message(recvd_message);
	  }
	}
      }

      if (s->force_ack > 0 && user->last_delivered >= user->last_ack_as_receiver + s->force_ack)
	lrm_send_ack(s, user);

      break;
    case LRM_ACK:
      if (len < LRM_MIN_PACKET_SIZE + sizeof(lrm_seqno)) {
	lrm_debug (("received ACK message with wrong data from %8x (%s, %u).\n", uid, address, seqno));
	return;

      }
      memcpy(&ack, buffer + LRM_MIN_PACKET_SIZE, sizeof(lrm_seqno));
      ack = ntohl(ack);
      if (ack > s->seqno || ack < user->last_ack_as_sender) {
	lrm_debug(("received bad ACK %d from %8x (%s).\n", ack, uid, address));
	return;
      }
      lrm_debug (("ACK %d from %8x (%s, %u).\n", ack, uid, address, seqno));
      user->last_ack_as_sender = ack;
      
      u = lrm_list_begin(s->users);
      while (u!=NULL) {
	if (((struct lrm_user *) lrm_list_content(u))->last_ack_as_sender < ack)
	  ack = ((struct lrm_user *) lrm_list_content(u))->last_ack_as_sender;
	u = lrm_list_next(u);
      }

      deleted = 0;
      m = lrm_list_begin(s->sentbufs);
      while (m!=NULL) {
	if (((struct lrm_sent_message *) lrm_list_content(m))->seqno <= ack) {
	  deleted = 1;
	  lrm_debug(("message %d was succesfully received. deleted.\n",
		     ((struct lrm_sent_message *) lrm_list_content(m))->seqno));
	  lrm_free_sent_message((struct lrm_sent_message *) lrm_list_content(m));
	  m = lrm_list_delete(s->sentbufs, m);
	} else
	  break;
      }
      if (deleted) {
	lrm_timer_resend (s);
	s->timer_resend_cur = s->timer_resend;
      }
    
      break;
    case LRM_PING:
      lrm_debug (("PING from %8x (%s, %u) ... PONG %d\n", uid, address, seqno, s->seqno-1));      
      message = lrm_create_message(s, LRM_PONG);
      lrm_send_message_ucast(s, addr, message, LRM_MIN_PACKET_SIZE);
      break;
    case LRM_PONG:
      lrm_debug (("PONG from %8x (%s, %u)\n", uid, address, seqno));
      user->ping = LRM_RESET_PING;
      break;
    case LRM_JOIN:
      lrm_debug (("JOIN from %8x (%s, %u) ... PONG %d\n", uid, address, seqno, s->seqno-1));
      if (user)
	lrm_del_user(s, uid);
      user = lrm_new_user(s, uid, addr);
      if (!user)
	lrm_memerr ("lrm_new_user");
      else
	user->last_delivered = user->last_ack_as_receiver = seqno;
      message = lrm_create_message(s, LRM_PONG);
      lrm_send_message_ucast(s, addr, message, LRM_MIN_PACKET_SIZE);
      break;
    case LRM_LEAVE:
      lrm_debug (("LEAVE from %8x (%s, %u)\n", uid, address, seqno));
      lrm_del_user(s, uid);
      break;
    default:
      lrm_debug(("received message with bad type %d from %8x (%s, %u).\n", type, uid, address, seqno));
  }
}


