/*
 * Light-weight Reliable Multicast
 *
 * lrm.h - Main lrm defines and structures.
 * Authors: Yoann Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifndef LRM_H
#define LRM_H

/* default values, timers in milliseconds  */
#define LRM_TIMER_SENDACK    300   
#define LRM_TIMER_RESEND     100
#define LRM_TIMER_PING       20000

#define LRM_WAIT_ANNOUNCE    10    /* in ping times */
#define LRM_WAIT_RESEND      8     /* in timer resend times */
#define LRM_MAX_PING         5     
#define LRM_MAX_RESEND       5
#define LRM_WAIT_FOR_RESENT  3     /* in ping times */
#define LRM_WIN_SIZE         50

typedef unsigned long   lrm_key;
typedef unsigned long   lrm_uid;
typedef unsigned short  lrm_type;
typedef unsigned long   lrm_seqno;

#define LRM_KEY              0x4c524d31 /* 'LRM1' */
#define LRM_RESET_PING       -1
#define LRM_SUCCESS          1
#define LRM_FAILURE          0
#define LRM_MAX_PACKET_SIZE  65536
#define LRM_MIN_PACKET_SIZE  (sizeof(lrm_key) + sizeof(lrm_uid) + sizeof(lrm_seqno) + sizeof(lrm_type))

#define LRM_MIN_TIMERES      10
#define LRM_MAX_TIMERES      30000

#ifdef __cplusplus
extern "C" { 
#endif

#ifdef LRM_DEBUG  
extern int lrm_debug_;

#define lrm_debug(format) if (lrm_debug_) { \
                            lrm_err ("[lrm %s:%d] ", __FILE__, __LINE__); \
			    lrm_err format; \
                          }
#else
#define lrm_debug(format) 
#endif

#ifdef LRM_MEM_DEBUG
typedef void*  *lrm_alloc_f(int);
void * talloc(int size, char * file, int line, lrm_alloc_f alloc);
#define malloc(a) talloc(a, __FILE__, __LINE__, malloc)
#define free(a) lrm_err("[lrm_mem %s:%d] free(%d)\n", __FILE__, __LINE__, (int) a); free(a);  
#endif
  
#define lrm_memerr(fonction) lrm_err ("lrm: memory allocation error in '%s' at %s:%d.\n", fonction, __FILE__, __LINE__)
  
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdarg.h>
#include "lrm_list.h"

struct lrm_session
{
  int sock;
  unsigned long add;
  unsigned short port;
  unsigned short ttl;
  lrm_list users;
  lrm_list sentbufs;
  lrm_list recvbufs;
  lrm_seqno seqno;
  lrm_uid uid;
#ifdef LRM_DEBUG
  float drop_probability;
#endif
  struct timeval timers_start;
  int win_size;
  int timer_sendack;
  int timer_resend;
  int timer_ping;
  int timer_sendack_cur;
  int timer_resend_cur;
  int timer_ping_cur;
  int wait_announce_cur;
  int wait_announce;
  int wait_resend;
  int max_ping;
  int max_resend;
  int wait_for_resent;
  int force_ack;
  int options;
};

struct lrm_sent_message {
  lrm_seqno seqno;
  void * data;
  unsigned int len;
  unsigned short resent;
  unsigned int wait;
};

struct lrm_recvd_message {
  lrm_seqno seqno;
  unsigned long addr;
  void * data;
  unsigned int len;
};

struct lrm_user {
  lrm_list messages;
  unsigned long addr;
  lrm_uid uid;
  lrm_seqno last_delivered;
  lrm_seqno last_ack_as_receiver;
  lrm_seqno last_ack_as_sender;
  short ping;
  unsigned int wait;
};

void lrm_err(const char * format, ...);
char * lrm_host(unsigned long addr);
char * lrm_addr(unsigned long addr);
void lrm_free_user(struct lrm_session *s, struct lrm_user *u);
struct lrm_user * lrm_new_user(struct lrm_session *s, lrm_uid uid, unsigned long addr);
struct lrm_user * lrm_get_user(struct lrm_session *s, lrm_uid uid);
void  lrm_del_user(struct lrm_session *s, lrm_uid uid);
void lrm_read_socket (struct lrm_session *s);

#ifdef __cplusplus
}
#endif

#endif
