/*
 * Light-weight Reliable Multicast
 *
 * lrm_api.c - Public lrm api routines.
 * Authors: Yoann Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#if STDC_HEADERS
#include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#else
extern int errno;
#endif
#include <stdlib.h>
#include <stdio.h>
#include "lrm_list.h"
#include "lrm.h"
#include "lrm_api.h"
#include "lrm_timers.h"
#include "lrm_messages.h"

struct lrm_session * lrm_create_session (unsigned long add, unsigned short port, unsigned short ttl, unsigned long uid) {
  int sock;
  char opt;
  struct ip_mreq mreq;
  struct sockaddr_in localExtremity;
  struct lrm_session * s;

  s = (struct lrm_session *) malloc(sizeof(struct lrm_session));
  if (!s) return NULL;
  
  sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock < 0) {
    lrm_err ("lrm: socket: %s\n", strerror(errno));
    return NULL;
  }
    
  localExtremity.sin_family = AF_INET;
  localExtremity.sin_addr.s_addr = INADDR_ANY;
  localExtremity.sin_port = port;

  if(bind(sock, (struct sockaddr *)&localExtremity, sizeof(localExtremity)) < 0) {
    lrm_err ("lrm: bind localExtremity.sin_addr = %s, " 
	     "localExtremity.sin_port = %hu : %s\n",
	      inet_ntoa(localExtremity.sin_addr), ntohs(port), strerror(errno));
    return NULL;
  } else {
    lrm_debug (("Ok: bind: localExtremity.sin_addr = %s, "
		"localExtremity.sin_port = %hu \n",
		inet_ntoa(localExtremity.sin_addr), ntohs(port)));
  }

  mreq.imr_interface.s_addr = INADDR_ANY;
  mreq.imr_multiaddr.s_addr = add; 
  if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &mreq, sizeof(mreq)) < 0) {
    lrm_err ("lrm: setsockopt: IPPROTO_IP, IP_ADD_MEMBERSHIP, "
	      "req=INADDR_ANY,%s: %s\n",inet_ntoa(mreq.imr_multiaddr), strerror(errno)); 
    return NULL;
  } else {
    lrm_debug (("Ok: setsockopt: IPPROTO_IP, IP_ADD_MEMBERSHIP, "
	      "req=INADDR_ANY,%s\n",inet_ntoa(mreq.imr_multiaddr))); 
  }

  opt = ttl;
  if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&opt, sizeof(opt)) < 0) {
    lrm_err("lrm: setsockopt: IPPROTO_IP, IP_MULTICAST_TTL, %d: %s\n",
	     opt, strerror(errno)); 
    return NULL; 
  } else {
    lrm_debug (("Ok: setsockopt: IPPROTO_IP, IP_MULTICAST_TTL, "
	      "ttl=%d\n",ttl));
  }
    
  opt = 0;
  if (setsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&opt, sizeof(opt)) < 0) {
    lrm_err ("lrm: setsockopt: IPPROTO_IP, IP_MULTICAST_LOOP, %d: %s\n",
	      opt, strerror(errno)); 
    return NULL; 
  } else {
    lrm_debug(("Ok: setsockopt: IPPROTO_IP, IP_MULTICAST_LOOP, "
	      "loop=%d\n",opt));
  }
  
  s->seqno = 1;
  s->uid = uid;
  s->sock = sock;
  s->add = add;
  s->ttl = ttl;
  s->port = port;
  s->users = lrm_list_new();
  if (!s->users)
    lrm_memerr ("lrm_list_new");
  s->sentbufs = lrm_list_new();
  if (!s->sentbufs)
    lrm_memerr ("lrm_list_new");
  s->recvbufs = lrm_list_new();
  if (!s->recvbufs)
    lrm_memerr ("lrm_list_new");

  s->timer_sendack = LRM_TIMER_SENDACK; 
  s->timer_resend = LRM_TIMER_RESEND;
  s->timer_ping = LRM_TIMER_PING;
  s->wait_announce = LRM_WAIT_ANNOUNCE;
  
  s->timer_sendack_cur = s->timer_sendack;
  s->timer_resend_cur = s->timer_resend;
  s->timer_ping_cur = s->timer_ping;
  s->wait_announce_cur = s->wait_announce;
  
  gettimeofday(&s->timers_start, NULL);

  s->wait_resend = LRM_WAIT_RESEND;
  s->max_ping = LRM_MAX_PING;
  s->max_resend = LRM_MAX_RESEND;
  s->wait_for_resent = LRM_WAIT_FOR_RESENT;
  s->win_size = LRM_WIN_SIZE;
  s->force_ack = s->win_size - 1;
  s->options = LRM_DEFAULT;

  lrm_debug(("Session ready for uid %8x.\n", uid));
  
  return s;
}

void lrm_destroy_session (struct lrm_session *s)
{
  lrm_iterator u;

  lrm_free_sent_messages(s);
  lrm_list_free(s->sentbufs);
  
  u=lrm_list_begin(s->users);
  while (u!=NULL) {
    lrm_free_user(s, (struct lrm_user *) lrm_list_content(u));
    u = lrm_list_next(u);
  }
  lrm_list_free(s->users);

  u = lrm_list_begin(s->recvbufs);
  while(u!=NULL) {
    lrm_free_recvd_message((struct lrm_recvd_message *) lrm_list_content(u));
    u = lrm_list_next(u);
  }
  lrm_list_free(s->recvbufs);

  close(s->sock);
  free(s);
}

unsigned long lrm_hostid() {
  unsigned long hostid;
  char hostname[128];
  struct hostent * h;
  
#ifdef HAVE_GETHOSTID
  hostid = gethostid();
  if (hostid)
    return hostid;
  lrm_err ("lrm: gethostid returned 0, reverting to manual method.\n");
#endif  
  if (gethostname(hostname, sizeof(hostname))) {
    lrm_err("lrm: gethostname: %s\n", strerror(errno));
    return 0;
  }      
  h = gethostbyname(hostname);
  if (h) {
    memcpy(&hostid, h->h_addr_list[0], sizeof(unsigned long));
    return hostid;
  }
  else
    return 0;
}

int lrm_can_send(struct lrm_session *s) {
  if (s->win_size > 0)
    return (lrm_list_count(s->sentbufs) < s->win_size);
  else
    return 1;
}

int lrm_send (struct lrm_session *s, void * data, int len) {
  int message_length;
  void * message;
  
  if (s->win_size>0 && (lrm_list_count(s->sentbufs) >= s->win_size)) {
    if (s->options & LRM_BLOCK)
      while (lrm_list_count(s->sentbufs) >= s->win_size)
	lrm_process_one_event(s);
    else
      return -1;
  }
  
  if (len > LRM_MAX_PACKET_SIZE - LRM_MIN_PACKET_SIZE - sizeof(lrm_seqno))
    len = LRM_MAX_PACKET_SIZE - LRM_MIN_PACKET_SIZE - sizeof(lrm_seqno);
  
  message = lrm_create_diffusion_message (s, s->seqno, data, len, &message_length);

  if (!lrm_list_isempty(s->users)) 
    if (lrm_save_sent_message (s, data, len) == LRM_FAILURE) {
      lrm_memerr ("lrm_save_sent_message");
      return -1;
    }
  
  s->seqno++;
  lrm_send_message_mcast(s, message, message_length);
  return len;
}

int lrm_can_recv(struct lrm_session *s) {
  return (!lrm_list_isempty(s->recvbufs));
}

int lrm_recv (struct lrm_session *s, void * data, int len) {
  return lrm_recvfrom(s, data, len, NULL, NULL);
}

int lrm_recvfrom(struct lrm_session *s, void * data, int len, struct sockaddr *from, int *fromlen) {
  struct sockaddr_in remoteExtremity;
  int remoteExtremityLen;
  struct lrm_recvd_message * m;
  lrm_iterator l;
  int mlen;

  if (lrm_list_isempty(s->recvbufs)) {
    if (s->options & LRM_BLOCK)
      while (lrm_list_isempty(s->recvbufs))
	lrm_process_one_event(s);
    else
      return -1;
  }

  l = lrm_list_begin(s->recvbufs);
  if (!l) return -1;

  m = (struct lrm_recvd_message *) lrm_list_content(l);

  if (from) {
    remoteExtremityLen = sizeof (remoteExtremity);
    remoteExtremity.sin_port = s->port;
    remoteExtremity.sin_family = AF_INET;
    remoteExtremity.sin_addr.s_addr= m->addr;
    if (fromlen) {
      if (*fromlen < remoteExtremityLen)
	remoteExtremityLen = *fromlen;
      else
	*fromlen = remoteExtremityLen;
    }
    memcpy((void *) from, (void *) &remoteExtremity, remoteExtremityLen);
  }

  mlen = m->len;
  if (len<mlen) mlen = len;
  if (mlen && m->data)
    memcpy(data, m->data, mlen);
  lrm_free_recvd_message(m);
  lrm_list_delete(s->recvbufs, l);
  return mlen;
}

void  lrm_join(struct lrm_session * s) {
  void * message;

  message = lrm_create_message (s, LRM_JOIN);
  lrm_send_message_mcast(s, message, LRM_MIN_PACKET_SIZE);
}

void  lrm_leave(struct lrm_session * s) {
  void * message;

  lrm_timer_sendack(s);
  message = lrm_create_message (s, LRM_LEAVE);
  lrm_send_message_mcast(s, message, LRM_MIN_PACKET_SIZE);
}

void  lrm_set_drop_probability(struct lrm_session *s, float percentage) {
#ifdef LRM_DEBUG
  s->drop_probability = percentage;
#else
  lrm_err("lrm: compile with LRM_DEBUG defined to be able to use drop_probability\n");
#endif
}

void  lrm_set_debug (int debug) {
#ifdef LRM_DEBUG
  lrm_debug_ = debug;
#else
  lrm_err("lrm: compile with LRM_DEBUG defined to be get debug messages\n");
#endif
}


int lrm_socket(struct lrm_session *s) {
  return s->sock;
}

int lrm_isidle(struct lrm_session *s) {
  return (lrm_list_isempty(s->sentbufs));
}

int lrm_set(struct lrm_session *s, int param, int value) {
  switch (param) {
  case LRM_SET_TIMER_SENDACK:
    s->timer_sendack = value;
    return LRM_SUCCESS;
  case LRM_SET_TIMER_RESEND:
    s->timer_resend = value;
    return LRM_SUCCESS;
  case LRM_SET_TIMER_PING:
    s->timer_ping = value;
    return LRM_SUCCESS;
  case LRM_SET_WAIT_RESEND:
    s->wait_resend = value;
    return LRM_SUCCESS;
  case LRM_SET_MAX_PING:
    s->max_ping = value;
    return LRM_SUCCESS;
  case LRM_SET_MAX_RESEND:
    s->max_resend = value;
    return LRM_SUCCESS;
  case LRM_SET_WAIT_FOR_RESENT:
    s->wait_for_resent = value;
    return LRM_SUCCESS;
  case LRM_SET_WIN_SIZE:
    s->win_size = value;
    return LRM_SUCCESS;
  case LRM_SET_FORCE_ACK:
    s->force_ack = value;
    return LRM_SUCCESS;
  case LRM_SET_WAIT_ANNOUNCE:
    s->wait_announce = value;
    return LRM_SUCCESS;
  case LRM_SET_OPTIONS:
    s->options = value;
    return LRM_SUCCESS;
  case LRM_GET_OPTIONS:
    return s->options;
  default:
    return LRM_FAILURE;
  }
}

int lrm_process_one_event(struct lrm_session *s) {
  fd_set fds;
  struct timeval timeout;
  int sel;
  
  FD_ZERO(&fds);
  FD_SET(s->sock, &fds);

  timeout = lrm_get_timeval(s);

  sel = select(s->sock+1, &fds, NULL, NULL, &timeout);

  if (sel<0) {
    lrm_process_timers(s);
    return sel;
  }

  if (sel>0 && FD_ISSET(s->sock, &fds))
    lrm_process_messages(s);

  lrm_process_timers(s);
  return 0;
}

void  lrm_process_timers(struct lrm_session *s) {
  struct timeval now;
  int offset;

  gettimeofday(&now, NULL);
  offset = (now.tv_sec - s->timers_start.tv_sec) * 1000
    + (now.tv_usec - s->timers_start.tv_usec) / 1000;

  if (!offset) return;

  s->timers_start = now;

  if ((s->timer_sendack_cur -= offset) < 0) {
    lrm_timer_sendack(s);
    s->timer_sendack_cur = s->timer_sendack;
  }

  if ((s->timer_ping_cur -= offset) < 0) {
    lrm_timer_ping(s);
    s->timer_ping_cur = s->timer_ping;
  }

  if ((s->timer_resend_cur -= offset) < 0) {
    lrm_timer_resend(s);
    s->timer_resend_cur = s->timer_resend;
  }
}

void  lrm_process_messages(struct lrm_session *s) {
  lrm_read_socket(s);
}

struct timeval lrm_get_timeval(struct lrm_session *s) {
  int milli;
  struct timeval tps;

  milli = LRM_MAX_TIMERES;
  if (s->timer_sendack_cur < milli) milli = s->timer_sendack_cur;
  if (s->timer_ping_cur < milli) milli = s->timer_ping_cur;
  if (s->timer_resend_cur < milli) milli = s->timer_resend_cur;
  if (milli < LRM_MIN_TIMERES) milli = LRM_MIN_TIMERES;

  tps.tv_sec = milli / 1000;
  tps.tv_usec = (milli % 1000) * 1000;
  return tps;
}
