/*
 * Light-weight Reliable Multicast
 *
 * lrm_list.c - Semi-generic double linked list.
 * Author: Gal Roualland.
 * See COPYING file for license information.
 */

#include <stdlib.h>
#include "lrm_list.h"

lrm_list lrm_list_new() {
  lrm_list lst;
  
  lst = (lrm_list_s *) malloc(sizeof(lrm_list_s));
  if (!lst) return NULL;
  
  lst->first = lst->last = NULL;
  lst->count = 0;
  
  return lst;
}

void lrm_list_free(lrm_list lst) {
  lrm_cell * l1, * l2;

  l1 = lst->first;
  while (l1) {
    l2 = l1->next;
    free(l1);
    l1 = l2;
  }

  free(lst);
}

#ifdef LRM_LIST_NO_MACROS
int lrm_list_isempty(lrm_list lst) {
  return ((lst->first==lst->last) && (lst->last==NULL));
}

lrm_iterator lrm_list_begin(lrm_list lst) {
  return lst->first;
}

lrm_iterator lrm_list_end(lrm_list lst) {
  return lst->last;
}

lrm_iterator lrm_list_next(lrm_iterator iter) {
  if (iter)
    return iter->next;
  else
    return NULL;
}

lrm_iterator lrm_list_previous(lrm_iterator iter) {
  if (iter)
    return iter->next;
  else
    return NULL;
}

void * lrm_list_content(lrm_iterator iter) {
  if (iter)
    return iter->data;
  else
    return NULL;
}

int lrm_list_count(lrm_list lst) {
  return lst->count;
}

int lrm_list_insert_first(lrm_list lst, void * data) {
  return lrm_list_insert_before(lst, lst->first, data);
}

int lrm_list_insert_last(lrm_list lst, void * data) {
  return lrm_list_insert_after(lst, lst->last, data);
}
#endif

int lrm_list_insert_before(lrm_list lst, lrm_iterator iter, void * data) {
  lrm_cell * c;

  c = (lrm_cell *) malloc(sizeof(lrm_cell));
  if (!c) return LRM_FAILURE;

  c->data = data;
  lst->count++;
  
  if (lrm_list_isempty(lst)) {
    c->previous = c->next = NULL;
    lst->first = lst->last = c;
    return LRM_SUCCESS;
  }
  
  if (!iter) {
    c->previous = lst->last;
    c->previous->next = c;
    c->next = NULL;
    lst->last = c;
    return LRM_SUCCESS;
  }

  c->previous = iter->previous;
  c->next = iter;
  c->next->previous = c;
  if (c->previous)
    c->previous->next = c;
  else
    lst->first = c;

  return LRM_SUCCESS;
}

int lrm_list_insert_after(lrm_list lst, lrm_iterator iter, void * data) {
  lrm_cell * c;

  c = (lrm_cell *) malloc(sizeof(lrm_cell));
  if (!c) return LRM_FAILURE;

  c->data = data;
  lst->count++;
  
  if (lrm_list_isempty(lst)) {
    c->previous = c->next = NULL;
    lst->first = lst->last = c;
    return LRM_SUCCESS;
  }
  
  if (!iter) {
    c->previous = lst->last;
    c->previous->next = c;
    c->next = NULL;
    lst->last = c;
    return LRM_SUCCESS;
  }

  c->previous = iter;
  c->next = iter->next;
  if (c->next)
    c->next->previous = c;
  else
    lst->last = c;
  c->previous->next = c;

  return LRM_SUCCESS;
}

lrm_iterator lrm_list_delete(lrm_list lst, lrm_iterator iter) {
  lrm_iterator ret;
  
  if (!iter) return NULL;

  if (iter->previous) 
    iter->previous->next = iter->next;
  else
    lst->first = iter->next;

  if (iter->next) {
    iter->next->previous = iter->previous;
    ret = iter->next;
  } else {
    lst->last = iter->previous;
    ret = NULL;
  }

  free(iter);
  lst->count--;
  
  return ret;
}

