/*
 * Light-weight Reliable Multicast
 *
 * lrm_messages.c - lrm messages handling code.
 * Authors: Yoan Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if STDC_HEADERS
#include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#else
extern int errno;
#endif
#include <stdlib.h>
#include <stdio.h>
#include "lrm_list.h"
#include "lrm.h"
#include "lrm_api.h"
#include "lrm_messages.h"

void * lrm_create_message(struct lrm_session *s, lrm_type stype) {
  static char message[LRM_MIN_PACKET_SIZE];
  lrm_key lrmkey = htonl(LRM_KEY);
  lrm_uid uid = htonl(s->uid);
  lrm_type type = htons(stype);
  lrm_seqno seqno = htonl(s->seqno - 1);

  memcpy ((void *) message,
	  (void *) &lrmkey, sizeof(lrm_key));
  memcpy ((void *) message + sizeof(lrm_key),
	  (void *) &uid, sizeof(lrm_uid));
  memcpy ((void *) message + sizeof(lrm_key) + sizeof(lrm_uid),
	  (void *) &type, sizeof(lrm_type));
  memcpy ((void *) message + sizeof(lrm_key) + sizeof(lrm_uid) + sizeof(lrm_type),
	  (void *) &seqno, sizeof(lrm_seqno));
  
  return (void *) message;
}

void * lrm_create_diffusion_message (struct lrm_session *s, lrm_seqno seqno, char * data, int data_length, int *message_length) {
  static char message[LRM_MAX_PACKET_SIZE];

  seqno = htonl(seqno);
  memcpy ((void *) message,
	  lrm_create_message(s, LRM_DIFFUSION), LRM_MIN_PACKET_SIZE);
  memcpy ((void *) message + LRM_MIN_PACKET_SIZE,
	  &seqno, sizeof(lrm_seqno));
  if (data_length)
    memcpy ((void *) message + LRM_MIN_PACKET_SIZE + sizeof(lrm_seqno),
	    (void *) data, data_length);

  *message_length = LRM_MIN_PACKET_SIZE + sizeof(lrm_seqno) + data_length;
  return (void *) message;
}

void *lrm_create_ack_message(struct lrm_session *s, lrm_seqno seqno, int *message_length) {
  static char message[LRM_MIN_PACKET_SIZE+sizeof(lrm_seqno)];

  seqno = htonl(seqno);
  memcpy ((void *) message,
	  lrm_create_message(s, LRM_ACK), LRM_MIN_PACKET_SIZE);
  memcpy ((void *) message + LRM_MIN_PACKET_SIZE,
	  &seqno, sizeof(lrm_seqno));
  
  *message_length = LRM_MIN_PACKET_SIZE + sizeof(lrm_seqno);
  return (void *) message;
}

void lrm_send_message_mcast(struct lrm_session *s, void * message, int message_length) 
{
  struct sockaddr_in remoteExtremity;
  int remoteExtremityLen;
  
  remoteExtremityLen = sizeof (remoteExtremity);
  remoteExtremity.sin_port = s->port;
  remoteExtremity.sin_family = AF_INET;
  remoteExtremity.sin_addr.s_addr= s->add;

  if (sendto(s->sock, message, message_length, 0,
	     (struct sockaddr *)&remoteExtremity, remoteExtremityLen)<0)
    lrm_err ("lrm: sendto: %s\n", strerror(errno));
}

void lrm_send_message_ucast(struct lrm_session *s, unsigned long addr, void * message, int message_length) 
{
  struct sockaddr_in remoteExtremity;
  int remoteExtremityLen;
  
  remoteExtremityLen = sizeof (remoteExtremity);
  remoteExtremity.sin_port = s->port;
  remoteExtremity.sin_family = AF_INET;
  remoteExtremity.sin_addr.s_addr= addr;
  
  if(sendto(s->sock, message, message_length, 0,
	    (struct sockaddr *)&remoteExtremity, sizeof (remoteExtremity))<0)
    lrm_err ("lrm: sendto: %s\n", strerror(errno));
}

void lrm_send_ack (struct lrm_session * s, struct lrm_user * u) {
  int message_length;
  void *message = lrm_create_ack_message (s, u->last_delivered, &message_length);
  lrm_debug(("sending ACK %d to uid %8x (%s).\n", u->last_delivered, u->uid,
	     lrm_addr(u->addr)));  
  lrm_send_message_ucast (s, u->addr, message, message_length);
  u->last_ack_as_receiver = u->last_delivered;
}

void lrm_give_message(struct lrm_session *s, struct lrm_user *u, struct lrm_recvd_message * m) {
  if (!lrm_list_insert_last(s->recvbufs, (void *) m))
    lrm_err("lrm: can't insert recvd message.\n");
  u->wait = s->wait_for_resent;
  if (u->last_delivered < m->seqno) u->last_delivered = m->seqno;
}

void  lrm_announce(struct lrm_session * s) {
  void * message = lrm_create_message (s, LRM_PING);

  lrm_debug(("PING through Multicast to announce ourself.\n"));
  lrm_send_message_mcast (s, message, LRM_MIN_PACKET_SIZE);
  s->wait_announce_cur = s->wait_announce;
}

void lrm_give_user_queue(struct lrm_session *s, struct lrm_user *user) {
  lrm_iterator m;
  struct lrm_recvd_message * msg;
  
  if (!lrm_list_isempty(user->messages)) {
    m = lrm_list_begin(user->messages);
    while (m!=NULL) {
      msg = (struct lrm_recvd_message *) lrm_list_content(m);
      if (msg->seqno == user->last_delivered+1) {
	lrm_give_message(s, user, msg);
	m = lrm_list_delete(user->messages, m);
      } else break;
    }
    if (user->last_ack_as_receiver != user->last_delivered)
      lrm_send_ack(s, user);
  }
}

int lrm_save_sent_message (struct lrm_session *s, void *data, int len) {
  struct lrm_sent_message *buffer =
    (struct lrm_sent_message *) malloc (sizeof (struct lrm_sent_message));

  if (!buffer) 
    return LRM_FAILURE;
  buffer->len = len;
  lrm_debug (("saved message %d (%d bytes), win %u/%u.\n", s->seqno, len, lrm_list_count(s->sentbufs)+1, s->win_size));
  buffer->seqno = s->seqno;
  if (len) {
    buffer->data = malloc(len);
    if (!buffer->data) {
      lrm_memerr("lrm_save_sent_message");
      return LRM_FAILURE;
    }
    memcpy (buffer->data, data, len);
  } else
    buffer->data = NULL;
  buffer->resent = 0;
  buffer->wait = s->wait_resend;
  return (lrm_list_insert_last(s->sentbufs, (void *) buffer));
}

void lrm_free_sent_message(struct lrm_sent_message * m) {
  if (m->data)
    free(m->data);
  free(m);
}

void lrm_free_sent_messages(struct lrm_session * s) {
  lrm_iterator m;
  
  m=lrm_list_begin(s->sentbufs);
  while (m!=NULL) {
    lrm_free_sent_message(((struct lrm_sent_message *) lrm_list_content(m)));
    m = lrm_list_delete(s->sentbufs, m);
  }
}

void lrm_free_recvd_message(struct lrm_recvd_message * m) {
  if (m->data)
    free(m->data);
  free(m);
}

void lrm_free_recvd_messages(struct lrm_session * s, struct lrm_user * u) {
  lrm_iterator m;
  
  m=lrm_list_begin(u->messages);
  while (m!=NULL) {
    if (s->options & LRM_DELIVER_ON_LEAVE)
      lrm_give_message(s, u, ((struct lrm_recvd_message *) lrm_list_content(m)));
    else
      lrm_free_recvd_message((struct lrm_recvd_message *) lrm_list_content(m));
    m = lrm_list_delete(u->messages, m);
  }
}















