/*
 * Light-weight Reliable Multicast
 *
 * lrm_timers.c - lrm timer events handling.
 * Authors: Yoan Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if STDC_HEADERS
#include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#else
extern int errno;
#endif
#include <stdlib.h>
#include <stdio.h>
#include "lrm_list.h"
#include "lrm.h"
#include "lrm_api.h"
#include "lrm_timers.h"
#include "lrm_messages.h"

void lrm_timer_sendack (struct lrm_session * s)
{
  lrm_iterator user_cell_ptr;

  user_cell_ptr = lrm_list_begin (s->users);
  while (user_cell_ptr != NULL) {
    struct lrm_user *user_ptr = (struct lrm_user *) lrm_list_content(user_cell_ptr);
    
    if (user_ptr->last_ack_as_receiver != user_ptr->last_delivered)
      lrm_send_ack(s, user_ptr);
    user_cell_ptr = lrm_list_next (user_cell_ptr);
  }
}

void lrm_timer_resend (struct lrm_session *s)
{
  lrm_iterator cell_ptr;
  int has_resent = 0;

  cell_ptr = lrm_list_begin (s->sentbufs);

  while (cell_ptr != NULL) {
    struct lrm_sent_message *buffer = (struct lrm_sent_message *) lrm_list_content(cell_ptr);
    
    if (buffer->wait) {
      buffer->wait--;
      cell_ptr = lrm_list_next(cell_ptr);
    } else {
      if (!has_resent) {
	void * message;
	int message_length;
	lrm_iterator cell_ptr_next;
	
	message = lrm_create_diffusion_message (s, buffer->seqno, buffer->data, buffer->len, &message_length);
	lrm_debug(("resending message %d (%d bytes) for %dth time.\n",
		   buffer->seqno, buffer->len, buffer->resent));
	lrm_send_message_mcast(s, message, message_length);
	(buffer->resent)++;
	cell_ptr_next = lrm_list_next(cell_ptr);
	if (buffer->resent > s->max_resend) {
	  lrm_debug (("message %d was never ack'ed after %d resent. deleting.\n",
		      buffer->seqno, s->max_resend));
	  lrm_free_sent_message((struct lrm_sent_message *) lrm_list_content(cell_ptr));
	  lrm_list_delete(s->sentbufs, cell_ptr);
	} else
	  buffer->wait = s->wait_resend;
	has_resent = 1;
	cell_ptr = cell_ptr_next;
      } else
	cell_ptr = lrm_list_next(cell_ptr);
    }
  }
}

void lrm_timer_ping (struct lrm_session * s)
{
  struct lrm_user *user_ptr;
  lrm_iterator user_cell_ptr;

  user_cell_ptr = lrm_list_begin(s->users);
  while (user_cell_ptr != NULL) {
    user_ptr = (struct lrm_user *) lrm_list_content(user_cell_ptr);
    if (user_ptr->wait) {
      user_ptr->wait--;
    } else {
      if (!lrm_list_isempty(user_ptr->messages)) {
	struct lrm_recvd_message * msg;
	msg = (struct lrm_recvd_message *)
	  lrm_list_content(user_ptr->messages->first);
	
	if (msg->seqno > user_ptr->last_delivered) {
	  lrm_debug (("LOST messages %d to %d from uid %8x (%s).\n",
		      user_ptr->last_delivered+1, msg->seqno - 1,
		      user_ptr->uid, lrm_addr(user_ptr->addr)));
	  user_ptr->last_ack_as_receiver = user_ptr->last_delivered = msg->seqno - 1;
	  lrm_give_user_queue(s, user_ptr);
	}
      }
      user_ptr->wait = s->wait_for_resent;
    }
    
    s->wait_announce_cur--;
    (user_ptr->ping)++;
    if (user_ptr->ping > 0)
      {
	if (user_ptr->ping > s->max_ping) {
	  lrm_iterator user_cell_ptr_next;
	  user_cell_ptr_next = lrm_list_next(user_cell_ptr);
	  lrm_free_user(s, user_ptr);
	  lrm_list_delete(s->users, user_cell_ptr);
	  user_cell_ptr = user_cell_ptr_next;
	} else {
	  if (s->wait_announce_cur!=0) {
	    void *message = lrm_create_message (s, LRM_PING);
	    lrm_send_message_ucast (s, user_ptr->addr, message, LRM_MIN_PACKET_SIZE);
	  }
	  user_cell_ptr = lrm_list_next (user_cell_ptr);
	}
      } else {
	user_cell_ptr = lrm_list_next (user_cell_ptr);
      }
  }

  if (s->wait_announce_cur == 0)
    lrm_announce(s);
}

