/*
 * Light-weight Reliable Multicast
 *
 * lrmcat.c - Simple lrm app, send what's on stdin, and display on stdout what's received.
 * Authors: Yoan Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
# include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#if STDC_HEADERS
#include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#include <stdlib.h>
#include <stdio.h>

#include "lrm_api.h"

#define BUFSIZE 4096

int goon = 1;
struct timeval startr;
struct timeval endr;
struct timeval starts;
struct timeval ends;
int starteds = 0, startedr = 0;

void usage(char * p) {
  fprintf(stderr,
	  "usage: %s [options] addr port\n"
	  "where valid options are :\n"
	  "  -d               sets debug mode\n"
	  "  -f percentage    sets drop probability 0-1 (debug)\n"
	  "  -a milliseconds  sets timer for sending ACKs\n"
	  "  -A n             sets number of packets before ACKs\n"
	  "  -r milliseconds  sets timer unit for resending packets\n"
	  "  -W milliseconds  sets number of timer to wait before resending.\n"
	  "  -G n             sets number of ping before giving up a packet\n" 
	  "  -B n             sets number of ping events before reannouncing\n"
	  "  -R n             sets max resend times for a packet\n"
	  "  -p milliseconds  sets ping frequency\n"
	  "  -P n             sets max pings before losing user\n"
	  "  -w n             sets send window size (0 to disable)\n"
	  "  -t n             sets ttl of multicast packets (defaults to 1)\n"
	  "  +/-s             sets/unsets strict behavior\n"
	  "  +/-b             sets/unsets block behavior\n"
	  "  +/-l             sets/unsets deliver on leave\n",
	  p);
  exit(EXIT_FAILURE);
}

RETSIGTYPE finish(int sig) {
  goon = 0;
}

int main(int argc, char ** argv) {
  unsigned char buf[BUFSIZE+1];
  fd_set fdsin;
  fd_set fdsout;
  int sel;
  int lu;
  struct timeval timeout;
  unsigned long addr;
  unsigned short port;
  short ttl = -1;
  u_int recvsize = 0;
  u_int sendsize = 0;
  struct hostent * host;
  struct lrm_session * lrm;
  double secr, secs;
  int input = 1;
  int arg;
  int dbg = -1, tack = -1, fack = -1, tres = -1, wres = -1, mres = -1, wfres = -1,
    tpng = -1, mpng = -1, win = -1, strict = -1, block = -1,
    deliv = -1, ann = -1;
  float fiab = 0;
  
  arg=1;

  if (argc<3) usage(argv[0]);

  while (arg<argc-2) {
    if (strlen(argv[arg])==2 && argv[arg][0]=='-') {
      switch (argv[arg][1]) {
      case 'd': dbg = 1; break;
      case 'f': fiab = atof(argv[++arg]); break;
      case 'a': tack = atoi(argv[++arg]); break;
      case 'A':	fack = atoi(argv[++arg]); break;
      case 'r':	tres = atoi(argv[++arg]); break;
      case 'W':	wres = atoi(argv[++arg]); break;
      case 'G':	wfres = atoi(argv[++arg]); break;
      case 'R':	mres = atoi(argv[++arg]); break;
      case 'p':	tpng = atoi(argv[++arg]); break;
      case 'P':	mpng = atoi(argv[++arg]); break;
      case 'w':	win = atoi(argv[++arg]); break;
      case 't':	ttl = atoi(argv[++arg]); break;
      case 'B':	ann = atoi(argv[++arg]); break;
      case 's': strict = 0; break;
      case 'b':	block = 0; break;
      case 'l':	deliv = 0; break;
      default: usage(argv[0]);
      }
    } else {
      if (strlen(argv[arg])==2 && argv[arg][0]=='+') {
	switch (argv[arg][1]) {
	case 's': strict = 1; break;
	case 'b': block = 1; break;
	case 'l': deliv = 1; break;
	default: usage(argv[0]);
	}
      } else usage(argv[0]);
    }
    arg++;
  }

  if (arg+2!=argc) usage(argv[0]);

  addr = inet_addr(argv[arg]);
  if(addr==-1) {
    host = gethostbyname(argv[arg]);
    if (!host) {
      fprintf(stderr, "%s: invalid address.\n", argv[0]);
      exit(EXIT_FAILURE);
    }
    memcpy(&addr, *host->h_addr_list, sizeof(addr));
  }

  port = atoi(argv[++arg]);
  if (port<1 || port>65535) {
    fprintf(stderr, "%s: invalid port.\n", argv[0]);
    exit(EXIT_FAILURE);
  }
  port = htons(port);

  if (dbg==1) lrm_set_debug(dbg);
  if (ttl<0) ttl = 1;

  lrm = lrm_create_session(addr, port, ttl, lrm_hostid());
  if (!lrm) {
    fprintf(stderr, "Can't create LRM socket");
    exit(EXIT_FAILURE);
  }

  if (fiab>=0) lrm_set_drop_probability(lrm, fiab);

  if (tack>0) lrm_set(lrm, LRM_SET_TIMER_SENDACK, tack);
  if (fack>-0) lrm_set(lrm, LRM_SET_FORCE_ACK, fack);
  if (tres>0) lrm_set(lrm, LRM_SET_TIMER_RESEND, tres);
  if (wres>0) lrm_set(lrm, LRM_SET_WAIT_RESEND, wres);
  if (wfres>0) lrm_set( lrm, LRM_SET_WAIT_FOR_RESENT, wfres);
  if (mres>0) lrm_set(lrm, LRM_SET_MAX_RESEND, mres);
  if (tpng>0) lrm_set(lrm, LRM_SET_TIMER_PING, tpng);
  if (mpng>0) lrm_set(lrm, LRM_SET_MAX_PING, mpng);
  if (win>=0) lrm_set(lrm, LRM_SET_WIN_SIZE, win);
  if (ann>=0) lrm_set(lrm, LRM_SET_WAIT_ANNOUNCE, ann);

  arg = lrm_set(lrm, LRM_GET_OPTIONS, 0);
  if (strict==1) arg |= LRM_STRICT; else if (strict==0) arg &= ~LRM_STRICT;
  if (block==1) arg |= LRM_BLOCK; else if (block==0) arg &= ~LRM_BLOCK;
  if (deliv==1) arg |= LRM_STRICT; else if (deliv==0) arg &= ~LRM_BLOCK;
  lrm_set(lrm, LRM_SET_OPTIONS, arg);
  
  signal(SIGINT, finish);
  lrm_join(lrm);
  sleep(2);

  while (goon) {
    FD_ZERO (&fdsin);
    FD_ZERO (&fdsout);

    FD_SET(lrm_socket(lrm), &fdsin);
    if (input)
      FD_SET (STDIN_FILENO, &fdsin);
    if (lrm_can_recv(lrm))
      FD_SET (STDOUT_FILENO, &fdsout);
    
    timeout = lrm_get_timeval(lrm);

    sel = select(lrm_socket(lrm)+1, &fdsin, &fdsout, NULL, &timeout);

    if (sel < 0) {
      perror("select");
      goon = 0;
      continue;
    }

    if (sel > 0) {
      if (input) {
	if(FD_ISSET(0, &fdsin)) {
	  if (!starteds) {
	    gettimeofday(&starts, NULL);
	    starteds = 1;
	  }
	  lu = read(0, buf+1, BUFSIZE);
	  sendsize += lu+1;
	  if (lu>0) {
	    gettimeofday(&ends, NULL);
	    lrm_send(lrm, buf, lu+1);
	  } else {
	    lrm_send(lrm, buf, 1);
	    input = 0;
	  }
	}
      }
      
      if (FD_ISSET(lrm_socket(lrm), &fdsin))
	lrm_process_messages(lrm);

      if (lrm_can_recv(lrm) && FD_ISSET(STDOUT_FILENO, &fdsout)) {
	lu = lrm_recv(lrm, buf, BUFSIZE+1);
	if (!startedr) {
	  gettimeofday(&startr, NULL);
	  startedr = 1;
	}
	if (lu>1) {
	  recvsize += lu;
	  write(STDOUT_FILENO, buf+1, lu-1);
	  gettimeofday(&endr, NULL);
	} else {
	  goon = 0;
	  recvsize++;
	}
      }
    }
    
    lrm_process_timers(lrm);
    
    if (!input && lrm_isidle(lrm)) goon = 0;   
  }
  
  lrm_leave(lrm);
  lrm_destroy_session(lrm);

  secs = ((double) ends.tv_sec - (double) starts.tv_sec) +  
    ((double) ends.tv_usec - (double) starts.tv_usec) / (double) 1000000;

  secr = ((double) endr.tv_sec - (double) startr.tv_sec) +  
    ((double) endr.tv_usec - (double) startr.tv_usec) / (double) 1000000;

  if (secs==0) secs = 0.1;
  if (secr==0) secr = 0.1;

  fprintf(stderr, "sent %d bytes [%g kb/s], received %d bytes [%g kb/s].\n",
	  sendsize, (double) sendsize / (secs * 1024),
	  recvsize, (double) recvsize / (secr * 1024));
  
  return(EXIT_SUCCESS);
  
}




