/*
 * Light-weight Reliable Multicast
 *
 * lrmtalk.c - Simple lrm usage example by implementing a multicasted talk.
 * Authors: Yoan Janvier, Gal Roualland.
 * See COPYING file for license information.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#if STDC_HEADERS
#include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
#include <stdio.h>
#include <stdlib.h>
#include "lrm_api.h"

#define BUFSIZE 4096
#define AUTSIZE 100

struct lrm_session * lrm;

char * getident() {
  struct passwd * pwd;
  static char aut[AUTSIZE];
  
  pwd = getpwuid(getuid());
  strcpy(aut, pwd->pw_name);
  aut[strlen(pwd->pw_name)]='@';
  gethostname(aut+strlen(pwd->pw_name)+1, AUTSIZE-(strlen(pwd->pw_name)+1));
  return aut;
}

RETSIGTYPE finish(int sig) {
  fprintf(stderr, "-- Leaving conference.\n");
  lrm_leave(lrm);
  lrm_destroy_session(lrm);
  exit(EXIT_SUCCESS);
}

void usage(char * p) {
  fprintf(stderr,
	  "usage: %s [options] addr port\n"
	  "where valid options are :\n"
	  "  -d               sets debug mode\n"
	  "  -f percentage    sets drop probability 0-1 (debug)\n"
	  "  -a milliseconds  sets timer for sending ACKs\n"
	  "  -A n             sets number of packets before ACKs\n"
	  "  -r milliseconds  sets timer unit for resending packets\n"
	  "  -W milliseconds  sets number of timer to wait before resending.\n"
	  "  -G n             sets number of ping before giving up a packet\n" 
	  "  -R n             sets max resend times for a packet\n"
	  "  -B n             sets number of ping events before reannouncing\n"
	  "  -p milliseconds  sets ping frequency\n"
	  "  -P n             sets max pings before losing user\n"
	  "  -w n             sets send window size (0 to disable)\n"
	  "  -t n             sets ttl of multicast packets (defaults to 1)\n"
	  "  +/-s             sets/unsets strict behavior\n"
	  "  +/-b             sets/unsets block behavior\n"
	  "  +/-l             sets/unsets deliver on leave\n",
	  p);
  exit(EXIT_FAILURE);
}

int main(int argc, char ** argv) {
  unsigned char buf[BUFSIZE];
  unsigned char bufs[BUFSIZE+AUTSIZE];
  int bufoffset;
  fd_set fdsin;
  int lu;
  struct timeval timeout;
  unsigned long addr;
  unsigned short port;
  short ttl = 1;
  int max;
  struct hostent * host;
  int arg = 1;
  int dbg = -1, tack = -1, fack = -1, tres = -1, wres = -1, mres = -1, wfres = -1,
    tpng = -1, mpng = -1, win = -1, strict = -1, block = -1,
    deliv = -1, ann = -1;
  float fiab = 0;
  
  printf("lrmtalk. LRM Test. Conferencing.\n\n");

  if (argc<3) usage(argv[0]);

  while (arg<argc-2) {
    if (strlen(argv[arg])==2 && argv[arg][0]=='-') {
      switch (argv[arg][1]) {
      case 'd': dbg = 1; break;
      case 'f': fiab = atof(argv[++arg]); break;
      case 'a': tack = atoi(argv[++arg]); break;
      case 'A':	fack = atoi(argv[++arg]); break;
      case 'r':	tres = atoi(argv[++arg]); break;
      case 'W':	wres = atoi(argv[++arg]); break;
      case 'G':	wfres = atoi(argv[++arg]); break;
      case 'R':	mres = atoi(argv[++arg]); break;
      case 'p':	tpng = atoi(argv[++arg]); break;
      case 'P':	mpng = atoi(argv[++arg]); break;
      case 'w':	win = atoi(argv[++arg]); break;
      case 't':	ttl = atoi(argv[++arg]); break;
      case 'B':	ann = atoi(argv[++arg]); break;
      case 's': strict = 0; break;
      case 'b':	block = 0; break;
      case 'l':	deliv = 0; break;
      default: usage(argv[0]);
      }
    } else {
      if (strlen(argv[arg])==2 && argv[arg][0]=='+') {
	switch (argv[arg][1]) {
	case 's': strict = 1; break;
	case 'b': block = 1; break;
	case 'l': deliv = 1; break;
	default: usage(argv[0]);
	}
      } else usage(argv[0]);
    }
    arg++;
  }

  if (arg+2!=argc) usage(argv[0]);

  addr = inet_addr(argv[arg]);
  if(addr==-1) {
    host = gethostbyname(argv[arg]);
    if (!host) {
      fprintf(stderr, "%s: invalid address.\n", argv[0]);
      exit(EXIT_FAILURE);
    }
    memcpy(&addr, *host->h_addr_list, sizeof(addr));
  }

  port = atoi(argv[++arg]);
  if (port<1 || port>65535) {
    fprintf(stderr, "%s: invalid port.\n", argv[0]);
    exit(EXIT_FAILURE);
  }
  port = htons(port);

  if (dbg==1) {
    fprintf(stderr, "-- Turning on debug in LRM.\n");
    lrm_set_debug(dbg);
  }
  if (ttl<0) ttl = 1;

  lrm = lrm_create_session(addr, port, ttl, lrm_hostid());
  if (!lrm) {
    fprintf(stderr, "Can't create LRM socket");
    exit(EXIT_FAILURE);
  }

  if (fiab>=0) {
    fprintf(stderr, "-- Setting fiability to %g.\n", fiab);
    lrm_set_drop_probability(lrm, fiab);
  }

  if (tack>0) lrm_set(lrm, LRM_SET_TIMER_SENDACK, tack);
  if (fack>-0) lrm_set(lrm, LRM_SET_FORCE_ACK, fack);
  if (tres>0) lrm_set(lrm, LRM_SET_TIMER_RESEND, tres);
  if (wres>0) lrm_set(lrm, LRM_SET_WAIT_RESEND, wres);
  if (wfres>0) lrm_set(lrm, LRM_SET_WAIT_FOR_RESENT, wfres);
  if (mres>0) lrm_set(lrm, LRM_SET_MAX_RESEND, mres);
  if (tpng>0) lrm_set(lrm, LRM_SET_TIMER_PING, tpng);
  if (mpng>0) lrm_set(lrm, LRM_SET_MAX_PING, mpng);
  if (win>=0) lrm_set(lrm, LRM_SET_WIN_SIZE, win);
  if (ann>=0) lrm_set(lrm, LRM_SET_WAIT_ANNOUNCE, ann);

  arg = lrm_set(lrm, LRM_GET_OPTIONS, 0);
  if (strict==1) arg |= LRM_STRICT; else if (strict==0) arg &= ~LRM_STRICT;
  if (block==1) arg |= LRM_BLOCK; else if (block==0) arg &= ~LRM_BLOCK;
  if (deliv==1) arg |= LRM_STRICT; else if (deliv==0) arg &= ~LRM_BLOCK;
  lrm_set(lrm, LRM_SET_OPTIONS, arg);

  fprintf(stderr, "-- Annoucing our arrival in conference.\n");
  lrm_join(lrm);
  sleep(2);
  strcpy(bufs, getident());
  bufoffset = strlen(bufs);
  bufs[bufoffset++]=':';
  bufs[bufoffset++]=' ';
  signal(SIGINT, finish);
  
  fprintf(stderr, "-- Entering select loop.\n");
  while (1) {
    FD_ZERO (&fdsin);
    FD_SET (0, &fdsin);
    FD_SET (lrm_socket(lrm), &fdsin);
    
    timeout = lrm_get_timeval(lrm);

    max = select(lrm_socket(lrm)+1, &fdsin, NULL, NULL, &timeout);

    if (max < 0) {
      perror("select");
      finish(0);
    }

    if (max > 0) {
      if(FD_ISSET(0, &fdsin)) {
	lu = read(0, buf, BUFSIZE);
	memcpy(bufs+bufoffset, buf, BUFSIZE);
	if (lu>0) lrm_send(lrm, bufs, lu+bufoffset);
      }
      
      if(FD_ISSET(lrm_socket(lrm), &fdsin)) {
	lrm_process_messages(lrm);
	while(lrm_can_recv(lrm)) {
	  lu = lrm_recv(lrm, buf, BUFSIZE);
	  if (lu>0) write(STDOUT_FILENO, buf, lu);
	}
      }
    }
    
    lrm_process_timers(lrm); 
  }  
}




